	/* G N O K K I
	   A Linux/Unix toolset and driver for the Nokia 3110/3810/8110 mobiles.
	   Copyright (C) Hugh Blemings, 1999  Released under the terms of 
       the GNU GPL, see file COPYING for more details.
	
	   This file:  fbus.c

	   Handles sending and receiving commands to/from the GSM handset
	   at present, tested on the Nokia 3810 only.  The file and 
	   associated routines are called FBUS (whatever) as I believe
	   (possibly wrongly) that this is what the protocol/interface
	   is called! :) */

#include	<termios.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<ctype.h>
#include	<sys/signal.h>
#include	<sys/types.h>
#include	<sys/time.h>

#include	"fbus.h"

	/* Global variables */
int						FBUS_LinkOK;

	/* Local variables */
enum FBUS_RX_States		State;
int						MessageLength;
u8						MessageBuffer[FBUS_MAX_RECEIVE_LENGTH];
unsigned char			MessageCSum;
int						BufferCount;
unsigned char			CalculatedCSum;
int						PortFD;
u8						RequestSequenceNumber;

	/* Initialise variables and state machine. */
void	FBUS_Initialise(int fd)
{
	State = FBUS_RX_Initialise;
	PortFD = fd;

	FBUS_LinkOK = false;
	RequestSequenceNumber = 0x10;

}

	/* State machine for receive handling.  Called once for each
	   character received from the handset/phone. */
void	FBUS_RX_StateMachine(char rx_byte)
{


	switch (State) {
		case FBUS_RX_Initialise:	
				
				BufferCount = 0;
				State = FBUS_RX_Sync;
				break;
	
					/* Messages from the phone start with an 0x04.  We
					   use this to "synchronise" with the incoming data
					   stream. */		
		case FBUS_RX_Sync:
				if (rx_byte == 0x04) {
					BufferCount = 0;
					CalculatedCSum = rx_byte;
					State = FBUS_RX_GetLength;
				}
				break;
		
					/* Next byte is the length of the message including
					   the message type byte but not including the checksum. */
		case FBUS_RX_GetLength:
				MessageLength = rx_byte;
				CalculatedCSum ^= rx_byte;
				State = FBUS_RX_GetMessage;
				break;

					/* Get each byte of the message.  We deliberately
					   get one too many bytes so we get the checksum
					   here as well. */
		case FBUS_RX_GetMessage:
				MessageBuffer[BufferCount] = rx_byte;
				BufferCount ++;

				if (BufferCount >= FBUS_MAX_RECEIVE_LENGTH) {
					State = FBUS_RX_Sync;		/* Should be PANIC */
				}
					/* If this is the last byte, it's the checksum */
				if (BufferCount >= MessageLength + 1) {

					MessageCSum = rx_byte;
						/* Compare against calculated checksum. */
					if (MessageCSum == CalculatedCSum) {
						/* Got checksum, matches calculated one so 
						   now pass to dispatch handler. */
						FBUS_RX_DispatchMessage();
						State = FBUS_RX_Sync;
					}
						/* Checksum didn't match so ignore. */
					else {
						fprintf(stdout, "CS Fail %02x != %02x", MessageCSum, CalculatedCSum);
						FBUS_RX_DisplayMessage();
						fflush(stdout);
						State = FBUS_RX_Sync;
					}
					
				}
				CalculatedCSum ^= rx_byte;
				break;

		default:
	}
}

	/* FBUS_RX_DispatchMessage
	   Once we've received a message from the phone, the command/message
	   type byte is used to call an appropriate handler routine or
	   simply acknowledge the message as required.  The rather verbose
	   names given to the handler routines reflect that the names
	   of the purpose of the handler are best guesses only hence both
	   the byte (0x0b) and the purpose (IncomingCall) are given. */
void	FBUS_RX_DispatchMessage(void)
{

		/* Switch on the basis of the message type byte */
	switch (MessageBuffer[0]) {
	
			/* 0x4a message is a response to our 0x4a request, assumed to
			   be a keepalive message of sorts.  No response required. */
		case 0x4a:	break;

			/* 0x4b messages are sent by handset in response (it seems)
			   to the keep alive packet.  We must acknowledge these it seems
			   by sending a response with the "sequence number" byte loaded
			   appropriately. */
		case 0x4b:	FBUS_RX_Handle0x4b_Status();
					break;

			/* 0x0b messages are sent by handset when an incoming call occurs,
			   this message must be acknowledged. */
		case 0x0b:	FBUS_RX_Handle0x0b_IncomingCall();
					break;

			/* Fairly self explanatory these two, though the outgoing 
			   call message has three (unexplained) data bytes. */
		case 0x0e:	FBUS_RX_Handle0x0e_OutgoingCallAnswered();
					break;

		case 0x0d:	FBUS_RX_Handle0x0d_IncomingCallAnswered();
					break;

			/* 0x10 messages are sent by the phone when an outgoing
			    call terminates. */
		case 0x10:	FBUS_RX_Handle0x10_EndOfOutgoingCall();
					break;

			/* 0x11 messages are sent by the phone when an incoming call
			   terminates.  There is some other data in the message, 
			   purpose as yet undertermined. */
		case 0x11: 	FBUS_RX_Handle0x11_EndOfIncomingCall();
					break;

			/* 0x12 messages are sent after the 0x10 message at the 
			   end of an outgoing call.  Significance of two messages
			   versus the one at the end of an incoming call  is as 
			   yet undertermined. */
		case 0x12: 	FBUS_RX_Handle0x12_EndOfOutgoingCall();
					break;

			/* 0x16 messages are sent by the phone during initialisation,
			   after the response to the 0x15 message.  Purpose is unclear, 
			   however the sequence bytes have been observed to change with 
			   differing software versions.
			   V06.61 (19/08/97) sends 0x10 0x02, V07.02 (17/03/98) sends 
			   0x30 0x02.  The actual data byte (0x02) is unchanged. 
			   Go figure :) */ 
		case 0x16:	FBUS_TX_SendStandardAcknowledge(0x16);
					break;

			/* We send an 0x3f message to the phone to request a different
			   type of status dump - this one seemingly concerned with 
			   SMS message centre details.  Phone responds with an ack to
			   our 0x3f request then sends an 0x41 message that has the
			   actual data in it. */
		case 0x3f:	break;	/* Don't send ack. */

			/* 0x41 Messages are sent in response to an 0x3f request. */
		case 0x41: 	FBUS_RX_Handle0x41_SMSMessageCentreData();
					break;

			/* 0x43 is a message we send to request data from a memory
			   location. The phone returns and acknowledge for the 
			   0x43 command then sends an 0x46 command with the actual data. */
		case 0x43:	break;
	
			/* 0x46 is sent after an 0x43 response with the requested data. */
		case 0x46: 	FBUS_RX_Handle0x46_MemoryLocationData();
					break;


			/* Here we  attempt to acknowledge and display messages we don't
			   understand fully... The phone will send the same message
			   several (5-6) times before giving up if no ack is received.
			   Phone also appears to refuse to send any of that message type
			   again until an init sequence is done again. */
		default:	if (FBUS_TX_SendStandardAcknowledge(MessageBuffer[0]) != true) {
						fprintf(stdout, "Standard Ack write failed!");
					}
						/* Now display unknown message to user. */
		 			FBUS_RX_DisplayMessage();
					break;
	}


}

	/* FBUS_RX_DisplayMessage
	   Called when a message we don't know about is received so that
	   the user can see what is going back and forth, and perhaps shed
	   some more light/explain another message type! */
	
void	FBUS_RX_DisplayMessage(void)
{
	int		count;
	int		line_count; 
	
	line_count = 0;
	fprintf(stdout, "Unknown: ");

	fprintf(stdout, "Msg Type: %02x ", MessageBuffer[0]);
	fprintf(stdout, "Msg Len: %02x ", MessageLength);
	fprintf(stdout, "Sequence Number: %02x ", MessageBuffer[1]);
	fprintf(stdout, "Checksum: %02x \n   ", MessageCSum);

	for (count = 2; count < MessageLength; count ++) {
		if (isprint(MessageBuffer[count])) {
			fprintf(stdout, "[%02x%c]", MessageBuffer[count], MessageBuffer[count]);
		}
		else {
			fprintf(stdout, "[%02x ]", MessageBuffer[count]);
		}
		line_count ++;

		if (line_count >= 8) {
			line_count = 0;
			fprintf(stdout, "\n   ");
		}
	}

	fprintf(stdout, "\n");
	fflush(stdout);

}

	/* FBUS_TX_UpdateSequenceNumber
	   Any command we originate must have a unique SequenceNumber.
	   Observation to date suggests that these values startx at 0x10
	   and cycle up to 0x17 before repeating again.  Perhaps more
	   accurately, the numbers cycle 0,1,2,3..7 with bit 4 of the byte
	   premanently set. */
void	FBUS_TX_UpdateSequenceNumber(void)
{
	RequestSequenceNumber ++;

	if (RequestSequenceNumber > 0x17 || RequestSequenceNumber < 0x10) {
		RequestSequenceNumber = 0x10;
	}
}

	/* Sends the "standard" acknowledge message back to the phone in
	   reponse to a message it sent automatically or in response to
	   a command sent to it.  The ack. algorithim isn't 100% understood
	   at this time. */
int		FBUS_TX_SendStandardAcknowledge(u8 message_type)
{
		/* Standard acknowledge seems to be to return an empty message
		   with the sequence number set to equal the sequence number
		   sent minus 0x08. */
	return(FBUS_TX_SendMessage(0, message_type, (MessageBuffer[1] & 0x1f) - 0x08, NULL));

}

	/* 0x4a messages appear to be a keepalive message and cause the phone
	   to send an 0x4a acknowledge and then an 0x4b message which has
	   status information in it. */
void 	FBUS_TX_Send0x4aMessage(void) 
{
	FBUS_TX_UpdateSequenceNumber();
	
	if (FBUS_TX_SendMessage(0, 0x4a, RequestSequenceNumber, NULL) != true) {
		fprintf(stdout, "Write failed!");	
	}
}

	/* 0x3f messages request SMS information, returned as an 0x41 message. */
void 	FBUS_TX_Send0x3fMessage(void) 
{
	FBUS_TX_UpdateSequenceNumber();
	
	if (FBUS_TX_SendMessage(0, 0x3f, RequestSequenceNumber, NULL) != true) {
		fprintf(stdout, "Write failed!");	
	}
}

	/* 0x43 messages appears to request the contents of a memory location.
	   The phone acknowledges the 0x43 message then returns an
	   0x46 message with the actual memory data.  As the 3810 only
	   has SIM memory, the memory_area value is untested and thus is
	   hardcoded to 0x02 which is the value observed.  Presumably the
	   8110 uses a different value to specify internal memory. */
void 	FBUS_TX_Send0x43_RequestMemoryLocation(u8 memory_area, u8 location) 
{
	u8		message[2];

	FBUS_TX_UpdateSequenceNumber();

	memory_area = 0x02;		/* For now */

		/* Build and send message */
	message[0] = memory_area;
	message[1] = location;

	if (FBUS_TX_SendMessage(2, 0x43, RequestSequenceNumber, message) != true) {
		fprintf(stdout, "Set Mem Loc Write failed!");	
	}

}

	/* 0x15 messages are sent by the pc during the initialisation phase
	   after sending lots of 0x55 characters [side note this may be some 
	   sort of autobaud sequence given the bit pattern generated
	   by sending 0x55].  Anyway, the contents of the message are not understood
	   so we simply send the same sequence observed between the W95 PC 
	   and the phone.  The init sequence may still be a bit flaky and is not
	   fully understood. */

void	FBUS_TX_Send0x15Message(u8 sequence_number)
{
	u8		message[20] = {0x02, 0x01, 0x07, 0xa2, 0x88, 0x81, 0x21, 0x55, 0x63, 0xa8, 0x00, 0x00, 0x07, 0xa3, 0xb8, 0x81, 0x20, 0x15, 0x63, 0x80};

	if (FBUS_TX_SendMessage(20, 0x15, sequence_number, message) != true) {
		fprintf(stdout, "Write failed!");	
	}
}

	/* Prepares the message header and sends it, prepends the
	   message start byte (0x01) and other values according
	   the value specified when called.  Calculates checksum
	   and then sends the lot down the pipe... */
int		FBUS_TX_SendMessage(u8 message_length, u8 message_type, u8 sequence_byte, u8 *buffer) 
{
	u8			out_buffer[FBUS_MAX_TRANSMIT_LENGTH + 5];
	int			count;
	unsigned char			checksum;

		/* Check message isn't too long, once the necessary
		   header and trailer bytes are included. */
	if ((message_length + 5) > FBUS_MAX_TRANSMIT_LENGTH) {

		return (false);
	}
		/* Now construct the message header. */
	out_buffer[0] = 0x01;	/* Start of message indicator */
	out_buffer[1] = message_length + 2;	/* Our message length refers to buffer, 
										   the value in the protcol includes 
										   type and sequence info. */
	out_buffer[2] = message_type;
	out_buffer[3] = sequence_byte;

		/* Copy in data if any. */	
	if (message_length != 0) {
		memcpy(out_buffer + 4, buffer, message_length);
	}
		/* Now calculate checksum over entire message 
		   and append to message. */
	checksum = 0;
	for (count = 0; count < message_length + 4; count ++) {
		checksum ^= out_buffer[count];
	}
	out_buffer[message_length + 4] = checksum;

		/* Send it out... */
	if (write(PortFD, out_buffer, message_length + 5) != message_length + 5) {
		return (false);
	}
	return (true);
}

void	FBUS_RX_Handle0x0b_IncomingCall(void)
{
	int		count;
	char	buffer[256];

		/* First, acknowledge message. */
	if (FBUS_TX_SendMessage(0, 0x0b, MessageBuffer[1] - 0x08, NULL) != true) {
		fprintf(stdout, "Write failed!");
	}

		/* Get info out of message.  At present, first three bytes are unknown
		   (though third seems to correspond to length of number).  Remaining 
		   bytes are the phone number, ASCII encoded. */
	fprintf(stdout, "Incoming call - status %02x %02x %02x, Number ", MessageBuffer[2], MessageBuffer[3], MessageBuffer[4]);

	for (count = 0; count < MessageBuffer[4]; count ++) {
		buffer[count] = MessageBuffer[5 + count];
	}
	buffer[count] = 0x00;

		/* Now display incoming call message. */
	fprintf(stdout, "%s.\n", buffer);
	fflush (stdout);
	
		
}

	/* 0x4b is a general status message. */
void	FBUS_RX_Handle0x4b_Status(void)
{
		/* First, send acknowledge. */
	if (FBUS_TX_SendStandardAcknowledge(0x4b) != true) {
		fprintf(stdout, "Write failed!");	
	}

		/* There are three data bytes in the status message, two have been
		   attributed to signal level, the third is presently unknown. 
		   Unknown byte hasb been observed to be 0x01 when connected to normal
		   network, 0x04 when no network available.   Steps through 0x02, 0x03
		   when incoming or outgoing calls occur...*/	
	fprintf(stdout, "Status: RF %02x, Batt %02x, Unknown byte is %02x.\n", MessageBuffer[3], MessageBuffer[4], MessageBuffer[2]);


	FBUS_LinkOK = true;

}

void	FBUS_RX_Handle0x10_EndOfOutgoingCall(void)
{
		/* As usual, acknowledge first. */
	if (FBUS_TX_SendMessage(0, 0x10, MessageBuffer[1] - 0x08, NULL) != true) {
		fprintf(stdout, "Write failed!");
	}
	fprintf(stdout, "Outgoing call terminated (0x10 message).\n");
	fflush(stdout);

}

void	FBUS_RX_Handle0x11_EndOfIncomingCall(void)
{
		/* As usual, acknowledge first. */
	if (FBUS_TX_SendMessage(0, 0x11, MessageBuffer[1] - 0x08, NULL) != true) {
		fprintf(stdout, "Write failed!");
	}
	fprintf(stdout, "Incoming call terminated.\n");
	fflush(stdout);

}

void	FBUS_RX_Handle0x12_EndOfOutgoingCall(void)
{
		/* As usual, acknowledge first. */
	if (FBUS_TX_SendMessage(0, 0x12, MessageBuffer[1] - 0x08, NULL) != true) {
		fprintf(stdout, "Write failed!");
	}
	fprintf(stdout, "Outgoing call terminated (0x12 message).\n");
	fflush(stdout);

}

void	FBUS_RX_Handle0x0d_IncomingCallAnswered(void)
{
		/* As usual, acknowledge first. */
	if (FBUS_TX_SendMessage(0, 0x0d, MessageBuffer[1] - 0x08, NULL) != true) {
		fprintf(stdout, "Write failed!");
	}
	fprintf(stdout, "Incoming call answered.\n");
	fflush(stdout);

}

void	FBUS_RX_Handle0x0e_OutgoingCallAnswered(void)
{
		/* As usual, acknowledge first. */
	if (FBUS_TX_SendMessage(0, 0x0e, MessageBuffer[1] - 0x08, NULL) != true) {
		fprintf(stdout, "Write failed!");
	}
	fprintf(stdout, "Outgoing call answered - status bytes %02x %02x %02x.\n", MessageBuffer[2], MessageBuffer[3], MessageBuffer[4]);
	fflush(stdout);

}

	/* Handle data contained in an 0x46 message which is sent by the phone
	   in reponse to a 0x43 message requesting memory contents. */
void	FBUS_RX_Handle0x46_MemoryLocationData(void)
{
	int		count;
	u8		label_length;	/* Stored in first data byte in message. */
	u8		number_length;	/* Stored immediately after label data. */

		/* As usual, acknowledge first. */
	if (FBUS_TX_SendMessage(0, 0x46, MessageBuffer[1] - 0x08, NULL) != true) {
		fprintf(stdout, "Write failed!");
	}

		/* Get/Calculate label and number length. */
	label_length = MessageBuffer[2];
	number_length = MessageBuffer[label_length + 3];

		/* And output data.  This is a dumbo way of outputting it but works for now... */	
	fprintf (stdout, "Memory Data: ");

	if (label_length == 0) {
		fprintf(stdout, "Label empty");
	}
	else {
		fprintf(stdout, "Label ");
		for (count = 0; count < label_length; count ++) {
			fprintf(stdout, "%c", MessageBuffer[count + 3]);
		}
	}

	if (number_length == 0) {
		fprintf(stdout, " Number empty");
	}
	else {
		fprintf(stdout, " Number ");
		for (count = 0; count < number_length; count ++) {
			fprintf(stdout, "%c", MessageBuffer[count + label_length + 4]);
		}
	}

	fprintf (stdout, "\n");

	fflush(stdout);
}

	/* Handle 0x41 message which is sent by phone in response to an
	   0x3f request.  Contains data about the Message Centre in use,
	   the only bit understood at this time is the phone number portion
	   at the end of the message.  This number appears to be null terminated
	   unlike others but we don't rely on it at this stage! */
void	FBUS_RX_Handle0x41_SMSMessageCentreData(void)
{
	u8		centre_number_length;
	int		count;
	
		/* As usual, acknowledge first. */
	if (!FBUS_TX_SendStandardAcknowledge(0x41)) {
		fprintf(stdout, "Write failed!");
	}

		/* Get Message Centre number length, which is byte 13 in message. */
	centre_number_length = MessageBuffer[13];

		/* First 12 bytes are status values, purpose unknown.  For now
		   simply display for user to mull over... */
	fprintf(stdout, "SMS Message Centre Data status bytes =");
	for (count = 0; count < 12; count ++) {
		fprintf(stdout, "0x%02x ", MessageBuffer[2 + count]);
	}

	if (centre_number_length == 0) {
		fprintf(stdout, "Number field emtpy.");
	}
	else {
		fprintf(stdout, "Number: ");
		for (count = 0; count < centre_number_length; count ++) {
			fprintf(stdout, "%c", MessageBuffer[14 + count]);
		}
	}

	fprintf(stdout, "\n");
	fflush(stdout);

}
